#version 430 core
//shader for environment map
out vec4 FragColor;

in vec3 fPosition;
in vec3 fNormal;
in vec2 fTextCoords;
in vec4 fPositionLightSpace;

uniform vec3 lampLightPosition;
uniform vec3 moonLightDirection;

uniform sampler2D fragTexture;
uniform sampler2D moonShadowMap;
uniform samplerCube lampShadowMap;

uniform vec3 viewPosition;
uniform float farPlane;
uniform vec3 specularObject;
uniform float shininessObject;


vec3 lampSamplePos[8] = vec3[]
(
   //vec3(1, 1,  1), vec3( 1, -1,  1), vec3(-1, -1,  1), vec3(-1, 1,  1), 
   //vec3(1, 1, -1), vec3( 1, -1, -1), vec3(-1, -1, -1), vec3(-1, 1, -1),
   //vec3(1, 1,  0), vec3( 1, -1,  0), vec3(-1, -1,  0), vec3(-1, 1,  0),
   vec3(1, 0,  1), vec3(-1,  0,  1), vec3( 1,  0, -1), vec3(-1, 0, -1),
   vec3(0, 1,  1), vec3( 0, -1,  1), vec3( 0, -1, -1), vec3( 0, 1, -1)
);

vec2 texSize = textureSize(moonShadowMap,0);
float x = 1.0/texSize.x;
float y = 1.0/texSize.y;

vec2 moonSamplePos[6] = vec2[](
		//vec2(-x,	y), 
        //vec2( 0.0f, y),
        //vec2( x,	y),
        vec2(-x,	0.0f),
        vec2( 0.0f, 0.0f),
        vec2( x,	0.0f),
        vec2(-x,	-y),
        vec2( 0.0f, -y),
        vec2( x,	-y)
);


vec3 calculateMoonLight();
vec3 calculateLampLight();
float calculateMoonShadow();
float calculateLampShadow();


void main()
{
	
	if((texture(fragTexture, fTextCoords)).a < 0.6) //discard semi transparent fragments
	discard;
	
	vec3 result = calculateLampLight();
	
	result += calculateMoonLight();

    FragColor = vec4(result, 1.0);	

    //FragColor = vec4(texture(fragTexture, fTextCoords).a, 0.0f, 0.0f, 1.0f); //debug alpha channel	   
} 

vec3 calculateLampLight()
{
	//phong lighting model
	//Ip = ka * ia + kd * (vec(surface to light source) * vec(surface normal)) * id + ks * (vec(reflection) * vec(dir to viewer))^shininess * is
	//i... light intensities, k... ambient, diffuse, specular components
	//all vectors normalized
	//vec(reflection) = 2 * (vec(surface to light source) * vec(surface normal)) * vec(surface normal) - vec(surface to light source)

	// ambient part
    float ambient = 0.15f; //ka
	//float ambient = 1.0f;
    vec3 ambientPart = ambient * texture(fragTexture, fTextCoords).rgb;
  	
    // diffuse part
	float diff = 0.6f; //kd
    vec3 norm = normalize(fNormal); //vec(surface normal)
    vec3 lightDirection = normalize(lampLightPosition - fPosition); //vec(surface to light source)
    float diffuse = max(dot(norm, lightDirection),0.0);	//diffuse can't be negative, even with 90degrees
    vec3 diffusePart = diff * diffuse * texture(fragTexture, fTextCoords).rgb;
    
    // specular part
    vec3 specular = vec3(0.3f, 0.3f, 0.3f); //id
    vec3 viewDir = normalize(viewPosition - fPosition); //vec(dir to viewer)
    vec3 reflectDir = reflect(-lightDirection, norm);  //vec(reflection)
    float spec = pow(max(dot(viewDir, reflectDir), 0.0), shininessObject);
    vec3 specularPart = specular * spec * specularObject;
    
	float lightDistance = length(lampLightPosition - fPosition); //(distance surface to light)

	//attenuation... object is less lit if farther away from light
	//att = 1 / (constant + linear factor * vec(distance surface to light) + quadratic factor * (distance surface to light)^2)
	float attenuation = 1.0 / (1.0 + 0.027 * lightDistance + 0.0028 * pow(lightDistance,2));

	 float shadowValue = calculateLampShadow();	 
	
	 return (ambientPart + (1.0 - shadowValue) * attenuation * (diffusePart + specularPart));
}

vec3 calculateMoonLight()
{
	//phong lighting model
	//Ip = ka * ia + kd * (vec(surface to light source) * vec(surface normal)) * id + ks * (vec(reflection) * vec(dir to viewer))^shininess * is
	//i... light intensities, k... ambient, diffuse, specular components
	//all vectors normalized
	//vec(reflection) = 2 * (vec(surface to light source) * vec(surface normal)) * vec(surface normal) - vec(surface to light source)


	// ambient part
    float ambient = 0.05f; //ka
	//float ambient = 1.0f; //ka, debug for shadows
	vec3 ambientPart = ambient * texture(fragTexture, fTextCoords).rgb;
	
	// diffuse part
	float diff = 0.05f; //kd
	vec3 norm = normalize(fNormal); //vec(surface normal)    
	vec3 lightDirection = normalize(-moonLightDirection);
    float diffuse = max(dot(norm, lightDirection),0.0);	//diffuse can't be negative, even with 90degrees
    vec3 diffusePart = diff * diffuse * texture(fragTexture, fTextCoords).rgb;

	 // specular part
	vec3 specular = vec3(0.2f, 0.2f, 0.2f); //id
    vec3 viewDir = normalize(viewPosition - fPosition); //vec(dir to viewer)
    vec3 reflectDir = reflect(-lightDirection, norm);  //vec(reflection)
    float spec = pow(max(dot(viewDir, reflectDir), 0.0), shininessObject);
    vec3 specularPart = specular * spec * specularObject;   
    
	float shadowValue = calculateMoonShadow();

	return (ambientPart + (1.0 - shadowValue) * (diffusePart + specularPart));	  
}


float calculateMoonShadow()
{	
	vec3 sampleCoordinates = fPositionLightSpace.xyz;	
	sampleCoordinates = sampleCoordinates * 0.5 + 0.5;  //bring from NDC space (range [-1, 1]) to range [0,1] for depth map 
   
    //get depth from shadow map
	//float shadowDepth = texture(moonShadowMap, sampleCoordinates.xy).r; //without pcf
    float fragmentDepth = sampleCoordinates.z;
	
	float bias = 0.003;	


	//PCF

	float shadow = 0.0;
	//also sample 8 neighboring shadow values
    for(int i = 0; i < 6; i++)
    {
        float shadowDepth = texture(moonShadowMap, sampleCoordinates.xy + moonSamplePos[i]).r;

		if(fragmentDepth - bias > shadowDepth) //fragment in shadow
		{
			shadow += 1.0;
		}
    }

	//average result of shadow calculation
	shadow = shadow/6.0;

	if(sampleCoordinates.z > 1.0)
        shadow = 0.0;
	
    return shadow;
}


float calculateLampShadow()
{
	//direction vector for sampling cube depth map
    vec3 lightDirection = fPosition - lampLightPosition;
    
    float shadowDepth = texture(lampShadowMap, lightDirection).r;
    
	//transform from [0,1] to [-1,1] to compare with fragment depth
    shadowDepth *= farPlane;
    
    float fragmentDepth = length(lightDirection); //fragment depth value	
    
    float bias = 0.2; //to prevent shadow acne	
    float shadow = 0.0;
	
	//if(fragmentDepth - bias > shadowDepth) 
		//shadow = 1.0;	   
	
	

	//pcf
	//float shadow = 0.0;
   
    int samples = 8;    
    float radius = 0.05;

    for(int i = 0; i < samples; ++i)
    {
        float shadowDepth = texture(lampShadowMap, lightDirection + lampSamplePos[i] * radius).r;
        shadowDepth *= farPlane;   //transform back from [0,1] to [-1,1] to compare with fragment depth

        if(fragmentDepth - bias > shadowDepth) //
		{
			shadow += 1.0;
		}
    }
    shadow /= float(samples);
	
    return shadow;
}




